# Compute the hcst, obs and fcst anomalies with or without cross-validation
# and return them, along with the hcst and obs climatologies.

compute_anomalies <- function(recipe, data, retrieve = TRUE) {

  ## TODO: Is this necessary?
  # Check
  if (retrieve) {
    if (is.null(recipe$Analysis$Workflow$Anomalies$compute)) {
      error(recipe$Run$logger, 
            paste("The anomaly module has been called, but the element",
                  "'Workflow:Anomalies:compute' is missing from the recipe."))
      stop()
    }
  }

  if (recipe$Analysis$Workflow$Anomalies$compute) {
    if (recipe$Analysis$Workflow$Anomalies$cross_validation) {
      cross <- TRUE
      cross_msg <- "with"
    } else {
      cross <- FALSE
      cross_msg <- "without"
    }
    original_dims <- dim(data$hcst$data)

    # Compute anomalies
    ## TODO: No cross-validation case
    arguments <- list(exp = data$hcst, obs = data$obs,
                      memb = TRUE, memb_dim = 'ensemble',
                      dat_dim = c('dat', 'ensemble'),
                      ncores = recipe$Analysis$ncores)
    if (inherits(data$hcst, "s2dv_cube")) {
      anom <- do.call(what = CST_Anomaly,
                      args = c(arguments, list(cross = cross,
                                               dim_anom = 'syear',
                                               ftime_dim = 'time')))
      anom$exp$data <- Reorder(anom$exp$data, names(original_dims))
      anom$obs$data <- Reorder(anom$obs$data, names(original_dims))
      # Change variable metadata
      for (var in data$hcst$attrs$Variable$varName) {
        # Change hcst longname
        data$hcst$attrs$Variable$metadata[[var]]$long_name <-
          paste(data$hcst$attrs$Variable$metadata[[var]]$long_name, "anomaly")
        # Change obs longname
        data$obs$attrs$Variable$metadata[[var]]$long_name <- 
        paste(data$obs$attrs$Variable$metadata[[var]]$long_name, "anomaly")
      }
    } else {
      anom <- do.call(what = Ano_CrossValid,
                      args = c(arguments, list(time_dim = 'syear')))
      ## TODO: Check if dim reorder is really necessary in all cases
      anom$exp <- Reorder(anom$exp, names(original_dims))
      anom$obs <- Reorder(anom$obs, names(original_dims))
    }
    # Save full fields
    hcst_fullvalue <- data$hcst
    obs_fullvalue <- data$obs
    # Save anomalies
    data$hcst <- anom$exp
    data$obs <- anom$obs
    remove(anom)
    
    # Compute forecast anomaly field
    if (!is.null(data$fcst)) {
      arguments <- list(time_dim = 'syear', dat_dim = c('dat', 'ensemble'),
                        memb = FALSE, memb_dim = 'ensemble', ftime_dim = 'time',
                        ncores = recipe$Analysis$ncores)
      if (inherits(data$fcst, "s2dv_cube")) {
        clim <- do.call(what = s2dv::Clim,
                        args = c(list(exp = hcst_fullvalue$data,
                                      obs = obs_fullvalue$data),
                                 arguments))
      } else {
        clim <- do.call(what = s2dv::Clim,
                        args = c(list(exp = hcst_fullvalue,
                                      obs = obs_fullvalue),
                                 arguments))
      }
      clim$clim_exp <- InsertDim(clim$clim_exp, posdim = 1, lendim = 1,
                                 name = "syear")
      # Store original dimensions
      dims <- dim(clim$clim_exp)
      ## Is there a better way to do this?
      # Repeat the array as many times as ensemble members
      clim$clim_exp <- rep(clim$clim_exp, data$fcst$dims[['ensemble']])
      # Rename and reorder dimensions
      dim(clim$clim_exp) <- c(dims, ensemble = data$fcst$dims[['ensemble']])
      clim$clim_exp <- Reorder(clim$clim_exp, order = names(dim(data$fcst$data)))
      # Get fcst anomalies
      if (inherits(data$fcst, "s2dv_cube")) {
        data$fcst$data <- data$fcst$data - clim$clim_exp
        # Change metadata
        for (var in data$fcst$attrs$Variable$varName) {
          data$fcst$attrs$Variable$metadata[[var]]$long_name <-
            paste(data$fcst$attrs$Variable$metadata[[var]]$long_name, "anomaly")
        }
      } else {
        data$fcst <- data$fcst - clim$clim_exp
      }
    }

    # Display success messages and save data
    if (retrieve) {     
      info(recipe$Run$logger,
           paste("The anomalies have been computed,", cross_msg,
                 "cross-validation. The original full fields are returned as",
                 "$hcst.full_val and $obs.full_val."))

      info(recipe$Run$logger, "##### ANOMALIES COMPUTED SUCCESSFULLY #####")
      # Save outputs
      recipe$Run$output_dir <- paste0(recipe$Run$output_dir,
                                      "/outputs/Anomalies/")
      # Save forecast
      if (recipe$Analysis$Workflow$Anomalies$save %in% 
          c('all', 'exp_only', 'fcst_only')) {
        save_forecast(recipe = recipe, data_cube = data$fcst, type = 'fcst')
      }
      # Save hindcast
      if (recipe$Analysis$Workflow$Anomalies$save %in%
          c('all', 'exp_only')) {
        save_forecast(recipe = recipe, data_cube = data$hcst, type = 'hcst')
      }
      # Save observation
      if (recipe$Analysis$Workflow$Anomalies$save == 'all') {
        save_observations(recipe = recipe, data_cube = data$obs)
      }
    }
  } else {
    warn(recipe$Run$logger, paste("The Anomalies module has been called, but",
         "recipe parameter Analysis:Variables:anomaly is set to FALSE.",
         "The full fields will be returned."))
    hcst_fullvalue <- NULL
    obs_fullvalue <- NULL
    info(recipe$Run$logger, "##### ANOMALIES NOT COMPUTED #####")
  }

  return(list(hcst = data$hcst, obs = data$obs, fcst = data$fcst,
              hcst.full_val = hcst_fullvalue, obs.full_val = obs_fullvalue))

}
